@@ -14,6 +14,7 @@
 #include <utility>
 #include <tuple>
 #include <algorithm>
+#include <iterator>
 
 #include "xexpression.hpp"
 #include "xiterator.hpp"
@@ -24,11 +25,48 @@ namespace xt
 
     namespace detail
     {
-        template <class... Args>
-        using common_size_type = std::common_type_t<typename Args::size_type...>;
+
+        /********************
+         * common_size_type *
+         ********************/
 
         template <class... Args>
-        using common_difference_type = std::common_type_t<typename Args::difference_type...>;
+        struct common_size_type_impl
+        {
+            using type = std::common_type_t<typename Args::size_type...>;
+        };
+
+        template <>
+        struct common_size_type_impl<>
+        {
+            using type = std::size_t;
+        };
+
+        template<class... Args>
+        using common_size_type = typename common_size_type_impl<Args...>::type;
+
+        /**************************
+         * common_difference type *
+         **************************/
+
+        template <class... Args>
+        struct common_difference_type_impl
+        {
+            using type = std::common_type_t<typename Args::difference_type...>;
+        };
+
+        template <>
+        struct common_difference_type_impl<>
+        {
+            using type = std::size_t;
+        };
+
+        template<class... Args>
+        using common_difference_type = typename common_difference_type_impl<Args...>::type;
+
+        /*********************
+         * common_value_type *
+         *********************/
 
         template <class... Args>
         using common_value_type = std::common_type_t<get_value_type<Args>...>;
@@ -61,78 +99,87 @@ namespace xt
 
     public:
 
-        using self_type = xfunction<F, E...>;
-        using functor_type = F;
+        using self_type = xfunction<F, R, E...>;
+        using functor_type = typename std::remove_reference<F>::type;
 
         using value_type = R;
         using reference = value_type;
         using const_reference = value_type;
         using pointer = value_type*;
         using const_pointer = const value_type*;
-        // Workaround for buggy error C2210 in VS2015
-        using size_type = std::common_type_t<typename E::size_type...>; // detail::common_size_type<E...>;
-        using difference_type = std::common_type_t<typename E::difference_type...>; //detail::common_difference_type<E...>;
+        using size_type = detail::common_size_type<E...>;
+        using difference_type = detail::common_difference_type<E...>;
 
-        using shape_type = std::vector<size_type>;
-        using strides_type = std::vector<size_type>;
+        using shape_type = promote_shape_t<typename E::shape_type...>;
         using closure_type = const self_type;
 
         using const_stepper = xfunction_stepper<F, R, E...>;
-        using const_iterator = xiterator<const_stepper>;
+        using const_iterator = xiterator<const_stepper, shape_type>;
         using const_storage_iterator = xf_storage_iterator<F, R, E...>;
 
         template <class Func>
         xfunction(Func&& f, const E&...e) noexcept;
 
-        size_type dimension() const;
-        shape_type shape() const;
+        size_type dimension() const noexcept;
+        const shape_type& shape() const;
 
         template <class... Args>
         const_reference operator()(Args... args) const;
 
         const_reference operator[](const xindex& index) const;
 
+        template <class It>
+        const_reference element(It first, It last) const;
+
         template <class S>
         bool broadcast_shape(S& shape) const;
 
         template <class S>
-        bool is_trivial_broadcast(const S& strides) const;
+        bool is_trivial_broadcast(const S& strides) const noexcept;
 
-        const_iterator begin() const;
-        const_iterator end() const;
-        const_iterator cbegin() const;
-        const_iterator cend() const;
-
-        const_iterator xbegin(const shape_type& shape) const;
-        const_iterator xend(const shape_type& shape) const;
-        const_iterator cxbegin(const shape_type& shape) const;
-        const_iterator cxend(const shape_type& shape) const;
+        const_iterator begin() const noexcept;
+        const_iterator end() const noexcept;
+        const_iterator cbegin() const noexcept;
+        const_iterator cend() const noexcept;
 
         template <class S>
-        const_stepper stepper_begin(const S& shape) const;
+        xiterator<const_stepper, S> xbegin(const S& shape) const noexcept;
+        template <class S>
+        xiterator<const_stepper, S> xend(const S& shape) const noexcept;
+        template <class S>
+        xiterator<const_stepper, S> cxbegin(const S& shape) const noexcept;
+        template <class S>
+        xiterator<const_stepper, S> cxend(const S& shape) const noexcept;
 
         template <class S>
-        const_stepper stepper_end(const S& shape) const;
+        const_stepper stepper_begin(const S& shape) const noexcept;
 
-        const_storage_iterator storage_begin() const;
-        const_storage_iterator storage_end() const;
+        template <class S>
+        const_stepper stepper_end(const S& shape) const noexcept;
+
+        const_storage_iterator storage_begin() const noexcept;
+        const_storage_iterator storage_end() const noexcept;
+
+        const_storage_iterator storage_cbegin() const noexcept;
+        const_storage_iterator storage_cend() const noexcept;
 
     private:
 
         template <std::size_t... I, class... Args>
         const_reference access_impl(std::index_sequence<I...>, Args... args) const;
 
-        template <std::size_t... I>
-        const_reference access_impl(const xindex& index, std::index_sequence<I...>) const;
+        template <std::size_t... I, class It>
+        const_reference element_access_impl(std::index_sequence<I...>, It first, It last) const;
 
         template <class Func, std::size_t... I>
-        const_stepper build_stepper(Func&& f, std::index_sequence<I...>) const;
+        const_stepper build_stepper(Func&& f, std::index_sequence<I...>) const noexcept;
 
         template <class Func, std::size_t... I>
-        const_storage_iterator build_storage_iterator(Func&& f, std::index_sequence<I...>) const;
+        const_storage_iterator build_storage_iterator(Func&& f, std::index_sequence<I...>) const noexcept;
 
         std::tuple<typename E::closure_type...> m_e;
-        typename std::remove_reference<F>::type m_f;
+        functor_type m_f;
+        shape_type m_shape;
 
         friend class xf_storage_iterator<F, R, E...>;
         friend class xfunction_stepper<F, R, E...>;
@@ -149,16 +196,17 @@ namespace xt
     public:
 
         using self_type = xf_storage_iterator<F, R, E...>;
+        using functor_type = typename std::remove_reference<F>::type;
         using xfunction_type = xfunction<F, R, E...>;
 
         using value_type = typename xfunction_type::value_type;
         using reference = typename xfunction_type::value_type;
         using pointer = typename xfunction_type::const_pointer;
         using difference_type = typename xfunction_type::difference_type;
-        using iterator_category = std::input_iterator_tag;
+        using iterator_category = std::forward_iterator_tag;
 
         template <class... It>
-        xf_storage_iterator(const xfunction_type* func, It&&... it);
+        xf_storage_iterator(const xfunction_type* func, It&&... it) noexcept;
 
         self_type& operator++();
         self_type operator++(int);
@@ -196,6 +244,7 @@ namespace xt
     public:
 
         using self_type = xfunction_stepper<F, R, E...>;
+        using functor_type = typename std::remove_reference<F>::type;
         using xfunction_type = xfunction<F, R, E...>;
 
         using value_type = typename xfunction_type::value_type;
@@ -208,7 +257,7 @@ namespace xt
         using shape_type = typename xfunction_type::shape_type;
 
         template <class... It>
-        xfunction_stepper(const xfunction_type* func, It&&... it);
+        xfunction_stepper(const xfunction_type* func, It&&... it) noexcept;
 
         void step(size_type dim, size_type n = 1);
         void step_back(size_type dim, size_type n = 1);
@@ -241,23 +290,6 @@ namespace xt
      * xfunction implementation *
      ****************************/
 
-    namespace detail
-    {
-        template <class E>
-        inline typename E::const_reference get_element(const E& e)
-        {
-            return e();
-        }
-
-        template <class E, class S, class... Args>
-        inline typename E::const_reference get_element(const E& e, S i, Args... args)
-        {
-            if(sizeof...(Args) >= e.dimension())
-                return get_element(e, args...);
-            return e(i, args...);
-        }
-    }
-
     /**
      * @name Constructor
      */
@@ -271,8 +303,9 @@ namespace xt
     template <class F, class R, class... E>
     template <class Func>
     inline xfunction<F, R, E...>::xfunction(Func&& f, const E&... e) noexcept
-        : m_f(std::forward<Func>(f)), m_e(e...)
+        : m_e(e...), m_f(std::forward<Func>(f)), m_shape(make_sequence<shape_type>(dimension(), size_type(1)))
     {
+        broadcast_shape(m_shape);
     }
     //@}
 
@@ -284,9 +317,9 @@ namespace xt
      * Returns the number of dimensions of the function.
      */
     template <class F, class R, class... E>
-    inline auto xfunction<F, R, E...>::dimension() const -> size_type
+    inline auto xfunction<F, R, E...>::dimension() const noexcept -> size_type
     {
-        auto func = [](size_type d, auto&& e) { return std::max(d, e.dimension()); };
+        auto func = [](size_type d, auto&& e) noexcept { return std::max(d, e.dimension()); };
         return accumulate(func, size_type(0), m_e);
     }
 
@@ -294,11 +327,9 @@ namespace xt
      * Returns the shape of the xfunction.
      */
     template <class F, class R, class... E>
-    inline auto xfunction<F, R, E...>::shape() const -> shape_type
+    inline auto xfunction<F, R, E...>::shape() const -> const shape_type&
     {
-        shape_type shape(dimension(), size_type(1));
-        broadcast_shape(shape);
-        return shape;
+        return m_shape;
     }
     //@}
 
@@ -321,9 +352,22 @@ namespace xt
     template <class F, class R, class... E>
     inline auto xfunction<F, R, E...>::operator[](const xindex& index) const -> const_reference
     {
-        return access_impl(index, std::make_index_sequence<sizeof...(E)>());
+        return element(index.cbegin(), index.cend());
+    }
+    
+    /**
+     * Returns a constant reference to the element at the specified position in the function.
+     * @param first iterator starting the sequence of indices
+     * @param second iterator starting the sequence of indices
+     * The number of indices in the squence should be equal or greater
+     * than the number of dimensions of the container.
+     */
+    template <class F, class R, class... E>
+    template <class It>
+    inline auto xfunction<F, R, E...>::element(It first, It last) const -> const_reference
+    {
+        return element_access_impl(std::make_index_sequence<sizeof...(E)>(), first, last);
     }
-
     //@}
     
     /**
@@ -333,7 +377,7 @@ namespace xt
     /**
      * Broadcast the shape of the function to the specified parameter.
      * @param shape the result shape
-     * @return a boolean indicating whether the broadcast is trivial
+     * @return a boolean indicating whether the broadcasting is trivial
      */
     template <class F, class R, class... E>
     template <class S>
@@ -345,13 +389,13 @@ namespace xt
     }
 
     /**
-     * Compares the specified strides with those of the container to see wether
-     * the broadcast is trivial.
-     * @return a boolean indicating whether the broadcast is trivial
+     * Compares the specified strides with those of the container to see whether
+     * the broadcasting is trivial.
+     * @return a boolean indicating whether the broadcasting is trivial
      */
     template <class F, class R, class... E>
     template <class S>
-    inline bool xfunction<F, R, E...>::is_trivial_broadcast(const S& strides) const
+    inline bool xfunction<F, R, E...>::is_trivial_broadcast(const S& strides) const noexcept
     {
         auto func = [&strides](bool b, auto&& e) { return b && e.is_trivial_broadcast(strides); };
         return accumulate(func, true, m_e);
@@ -366,11 +410,9 @@ namespace xt
      * Returns a constant iterator to the first element of the function.
      */
     template <class F, class R, class... E>
-    inline auto xfunction<F, R, E...>::begin() const -> const_iterator
+    inline auto xfunction<F, R, E...>::begin() const noexcept -> const_iterator
     {
-        shape_type shape(dimension(), size_type(1));
-        broadcast_shape(shape);
-        return xbegin(shape);
+        return xbegin(shape());
     }
 
     /**
@@ -378,18 +420,16 @@ namespace xt
      * of the function.
      */
     template <class F, class R, class... E>
-    inline auto xfunction<F, R, E...>::end() const -> const_iterator
+    inline auto xfunction<F, R, E...>::end() const noexcept -> const_iterator
     {
-        shape_type shape(dimension(), size_type(1));
-        broadcast_shape(shape);
-        return xend(shape);
+        return xend(shape());
     }
 
     /**
      * Returns a constant iterator to the first element of the function.
      */
     template <class F, class R, class... E>
-    inline auto xfunction<F, R, E...>::cbegin() const -> const_iterator
+    inline auto xfunction<F, R, E...>::cbegin() const noexcept -> const_iterator
     {
         return begin();
     }
@@ -399,7 +439,7 @@ namespace xt
      * of the function.
      */
     template <class F, class R, class... E>
-    inline auto xfunction<F, R, E...>::cend() const -> const_iterator
+    inline auto xfunction<F, R, E...>::cend() const noexcept -> const_iterator
     {
         return end();
     }
@@ -410,9 +450,10 @@ namespace xt
      * @param shape the shape used for braodcasting
      */
     template <class F, class R, class... E>
-    inline auto xfunction<F, R, E...>::xbegin(const shape_type& shape) const -> const_iterator
+    template <class S>
+    inline auto xfunction<F, R, E...>::xbegin(const S& shape) const noexcept -> xiterator<const_stepper, S>
     {
-        return const_iterator(stepper_begin(shape), shape);
+        return xiterator<const_stepper, S>(stepper_begin(shape), shape);
     }
 
     /**
@@ -421,9 +462,10 @@ namespace xt
      * @param shape the shape used for broadcasting
      */
     template <class F, class R, class... E>
-    inline auto xfunction<F, R, E...>::xend(const shape_type& shape) const -> const_iterator
+    template <class S>
+    inline auto xfunction<F, R, E...>::xend(const S& shape) const noexcept -> xiterator<const_stepper, S>
     {
-        return const_iterator(stepper_end(shape), shape);
+        return xiterator<const_stepper, S>(stepper_end(shape), shape);
     }
 
     /**
@@ -432,7 +474,8 @@ namespace xt
      * @param shape the shape used for braodcasting
      */
     template <class F, class R, class... E>
-    inline auto xfunction<F, R, E...>::cxbegin(const shape_type& shape) const -> const_iterator
+    template <class S>
+    inline auto xfunction<F, R, E...>::cxbegin(const S& shape) const noexcept -> xiterator<const_stepper, S>
     {
         return xbegin(shape);
     }
@@ -443,7 +486,8 @@ namespace xt
      * @param shape the shape used for broadcasting
      */
     template <class F, class R, class... E>
-    inline auto xfunction<F, R, E...>::cxend(const shape_type& shape) const -> const_iterator
+    template <class S>
+    inline auto xfunction<F, R, E...>::cxend(const S& shape) const noexcept -> xiterator<const_stepper, S>
     {
         return xend(shape);
     }
@@ -451,17 +495,17 @@ namespace xt
 
     template <class F, class R, class... E>
     template <class S>
-    inline auto xfunction<F, R, E...>::stepper_begin(const S& shape) const -> const_stepper
+    inline auto xfunction<F, R, E...>::stepper_begin(const S& shape) const noexcept -> const_stepper
     {
-        auto f = [&shape](const auto& e) { return e.stepper_begin(shape); };
+        auto f = [&shape](const auto& e) noexcept { return e.stepper_begin(shape); };
         return build_stepper(f, std::make_index_sequence<sizeof...(E)>());
     }
 
     template <class F, class R, class... E>
     template <class S>
-    inline auto xfunction<F, R, E...>::stepper_end(const S& shape) const -> const_stepper
+    inline auto xfunction<F, R, E...>::stepper_end(const S& shape) const noexcept -> const_stepper
     {
-        auto f = [&shape](const auto& e) { return e.stepper_end(shape); };
+        auto f = [&shape](const auto& e) noexcept { return e.stepper_end(shape); };
         return build_stepper(f, std::make_index_sequence<sizeof...(E)>());
     }
 
@@ -469,13 +513,13 @@ namespace xt
      * @name Storage iterators
      */
     /**
-     * Returns a constant iterator to the first element of the buffer
+     * Returns an iterator to the first element of the buffer
      * containing the elements of the function.
      */
     template <class F, class R, class... E>
-    inline auto xfunction<F, R, E...>::storage_begin() const -> const_storage_iterator
+    inline auto xfunction<F, R, E...>::storage_begin() const noexcept -> const_storage_iterator
     {
-        auto f = [](const auto& e) { return e.storage_begin(); };
+        auto f = [](const auto& e) noexcept { return e.storage_cbegin(); };
         return build_storage_iterator(f, std::make_index_sequence<sizeof...(E)>());
     }
 
@@ -484,11 +528,31 @@ namespace xt
      * element of the buffer containing the elements of the function.
      */
     template <class F, class R, class... E>
-    inline auto xfunction<F, R, E...>::storage_end() const -> const_storage_iterator
+    inline auto xfunction<F, R, E...>::storage_end() const noexcept -> const_storage_iterator
     {
-        auto f = [](const auto& e) { return e.storage_end(); };
+        auto f = [](const auto& e) noexcept { return e.storage_cend(); };
         return build_storage_iterator(f, std::make_index_sequence<sizeof...(E)>());
     }
+
+    /**
+     * Returns a constant iterator to the first element of the buffer
+     * containing the elements of the function.
+     */
+    template <class F, class R, class... E>
+    inline auto xfunction<F, R, E...>::storage_cbegin() const noexcept -> const_storage_iterator
+    {
+        return storage_begin();
+    }
+
+    /**
+     * Returns a constant iterator to the element following the last
+     * element of the buffer containing the elements of the function.
+     */
+    template <class F, class R, class... E>
+    inline auto xfunction<F, R, E...>::storage_cend() const noexcept -> const_storage_iterator
+    {
+        return storage_end();
+    }
     //@}
 
     template <class F, class R, class... E>
@@ -499,22 +563,22 @@ namespace xt
     }
 
     template <class F, class R, class... E>
-    template <std::size_t... I>
-    inline auto xfunction<F, R, E...>::access_impl(const xindex& index, std::index_sequence<I...>) const -> const_reference
+    template <std::size_t... I, class It>
+    inline auto xfunction<F, R, E...>::element_access_impl(std::index_sequence<I...>, It first, It last) const -> const_reference
     {
-        return m_f(std::get<I>(m_e)[index]...);
+        return m_f((std::get<I>(m_e).element(first, last))...);
     }
 
     template <class F, class R, class... E>
     template <class Func, std::size_t... I>
-    inline auto xfunction<F, R, E...>::build_stepper(Func&& f, std::index_sequence<I...>) const -> const_stepper
+    inline auto xfunction<F, R, E...>::build_stepper(Func&& f, std::index_sequence<I...>) const noexcept -> const_stepper
     {
         return const_stepper(this, f(std::get<I>(m_e))...);
     }
 
     template <class F, class R, class... E>
     template <class Func, std::size_t... I>
-    inline auto xfunction<F, R, E...>::build_storage_iterator(Func&& f, std::index_sequence<I...>) const -> const_storage_iterator
+    inline auto xfunction<F, R, E...>::build_storage_iterator(Func&& f, std::index_sequence<I...>) const noexcept -> const_storage_iterator
     {
         return const_storage_iterator(this, f(std::get<I>(m_e))...);
     }
@@ -525,7 +589,7 @@ namespace xt
 
     template <class F, class R, class... E>
     template <class... It>
-    inline xf_storage_iterator<F, R, E...>::xf_storage_iterator(const xfunction_type* func, It&&... it)
+    inline xf_storage_iterator<F, R, E...>::xf_storage_iterator(const xfunction_type* func, It&&... it) noexcept
         : p_f(func), m_it(std::forward<It>(it)...)
     {
     }
@@ -586,7 +650,7 @@ namespace xt
 
     template <class F, class R, class... E>
     template <class... It>
-    inline xfunction_stepper<F, R, E...>::xfunction_stepper(const xfunction_type* func, It&&... it)
+    inline xfunction_stepper<F, R, E...>::xfunction_stepper(const xfunction_type* func, It&&... it) noexcept
         : p_f(func), m_it(std::forward<It>(it)...)
     {
     }
