@@ -9,13 +9,16 @@
 #ifndef XVIEW_HPP
 #define XVIEW_HPP
 
+#include <cstddef>
 #include <utility>
 #include <type_traits>
 #include <tuple>
+#include <array>
 #include <algorithm>
 
 #include "xarray.hpp"
 #include "xslice.hpp"
+#include "xiterator.hpp"
 
 namespace xt
 {
@@ -24,7 +27,7 @@ namespace xt
      * xview declaration *
      *********************/
 
-    template <class E, class... S>
+    template <bool is_const, class E, class... S>
     class xview_stepper;
 
     template <class E, class... S>
@@ -36,6 +39,9 @@ namespace xt
         using temporary_type = xarray<typename E::value_type>;
     };
 
+    template <class ST, class... S>
+    struct xview_shape_type;
+
     /**
      * @class xview
      * @brief Multidimensional view with tensor semantic.
@@ -67,12 +73,12 @@ namespace xt
         using size_type = typename E::size_type;
         using difference_type = typename E::difference_type;
 
-        using shape_type = std::vector<size_type>;
-        using strides_type = std::vector<size_type>;
+        using shape_type = typename xview_shape_type<typename E::shape_type, S...>::type;
+
         using slice_type = std::tuple<S...>;
 
-        using stepper = xview_stepper<E, S...>;
-        using const_stepper = xview_stepper<const E, S...>;
+        using stepper = xview_stepper<false, E, S...>;
+        using const_stepper = xview_stepper<true, E, S...>;
 
         using iterator = xiterator<stepper, shape_type>;
         using const_iterator = xiterator<const_stepper, shape_type>;
@@ -95,9 +101,15 @@ namespace xt
 
         template <class... Args>
         reference operator()(Args... args);
+        reference operator[](const xindex& index);
+        template <class It>
+        reference element(It first, It last);
 
         template <class... Args>
         const_reference operator()(Args... args) const;
+        const_reference operator[](const xindex& index) const;
+        template <class It>
+        const_reference element(It first, It last) const;
 
         template <class ST>
         bool broadcast_shape(ST& shape) const;
@@ -113,13 +125,19 @@ namespace xt
         const_iterator cbegin() const;
         const_iterator cend() const;
 
-        iterator xbegin(const shape_type& shape);
-        iterator xend(const shape_type& shape);
+        template <class ST>
+        xiterator<stepper, ST> xbegin(const ST& shape);
+        template <class ST>
+        xiterator<stepper, ST> xend(const ST& shape);
 
-        const_iterator xbegin(const shape_type& shape) const;
-        const_iterator xend(const shape_type& shape) const;
-        const_iterator cxbegin(const shape_type& shape) const;
-        const_iterator cxend(const shape_type& shape) const;
+        template <class ST>
+        xiterator<const_stepper, ST> xbegin(const ST& shape) const;
+        template <class ST>
+        xiterator<const_stepper, ST> xend(const ST& shape) const;
+        template <class ST>
+        xiterator<const_stepper, ST> cxbegin(const ST& shape) const;
+        template <class ST>
+        xiterator<const_stepper, ST> cxend(const ST& shape) const;
 
         template <class ST>
         stepper stepper_begin(const ST& shape);
@@ -137,6 +155,9 @@ namespace xt
         const_storage_iterator storage_begin() const;
         const_storage_iterator storage_end() const;
 
+        const_storage_iterator storage_cbegin() const;
+        const_storage_iterator storage_cend() const;
+
     private:
 
         E& m_e;
@@ -155,13 +176,18 @@ namespace xt
         template <size_type I, class... Args>
         std::enable_if_t<(I >= sizeof...(S)), size_type> index(Args... args) const;
 
-        template<size_type I, class T, class... Args>
-        size_type sliced_access(const xslice<T>& slice, Args... args) const;
+        template<size_type I, class T>
+        size_type sliced_access(const xslice<T>& slice) const;
+
+        template<size_type I, class T, class Arg, class... Args>
+        size_type sliced_access(const xslice<T>& slice, Arg arg, Args... args) const;
 
         template<size_type I, class T, class... Args>
         disable_xslice<T, size_type> sliced_access(const T& squeeze, Args...) const;
 
         using temporary_type = typename xcontainer_inner_types<self_type>::temporary_type;
+        using base_index_type = get_index_type<shape_type>;
+
         void assign_temporary_impl(temporary_type& tmp);
 
         friend class xview_semantic<xview<E, S...>>;
@@ -194,14 +220,14 @@ namespace xt
     template <class V>
     using get_stepper = typename detail::get_stepper_impl<V>::type;
 
-    template <class E, class... S>
+    template <bool is_const, class E, class... S>
     class xview_stepper
     {
 
     public:
 
-        using view_type = std::conditional_t<std::is_const<E>::value,
-                                             const xview<std::remove_const_t<E>, S...>,
+        using view_type = std::conditional_t<is_const,
+                                             const xview<E, S...>,
                                              xview<E, S...>>;
         using substepper_type = get_stepper<view_type>;
 
@@ -233,13 +259,13 @@ namespace xt
         size_type m_offset;
     };
 
-    template <class E, class... S>
-    bool operator==(const xview_stepper<E, S...>& lhs,
-                    const xview_stepper<E, S...>& rhs);
+    template <bool is_const, class E, class... S>
+    bool operator==(const xview_stepper<is_const, E, S...>& lhs,
+                    const xview_stepper<is_const, E, S...>& rhs);
 
-    template <class E, class... S>
-    bool operator!=(const xview_stepper<E, S...>& lhs,
-                    const xview_stepper<E, S...>& rhs);
+    template <bool is_const, class E, class... S>
+    bool operator!=(const xview_stepper<is_const, E, S...>& lhs,
+                    const xview_stepper<is_const, E, S...>& rhs);
 
     /********************************
      * helper functions declaration *
@@ -257,6 +283,32 @@ namespace xt
     template <class... S>
     constexpr std::size_t integral_skip(std::size_t i);
 
+    // return slice evaluation and increment iterator
+    template <class S, class It>
+    inline disable_xslice<S, std::size_t> get_slice_value(const S& s, It&) noexcept
+    {
+        return s;
+    };
+
+    template <class S, class It>
+    inline auto get_slice_value(const xslice<S>& slice, It& it) noexcept
+    {
+        return slice.derived_cast()(*it++);
+    };
+
+    // meta-function returning the shape type for an xview 
+    template <class ST, class... S>
+    struct xview_shape_type
+    {
+        using type = ST;
+    };
+
+    template <class I, std::size_t L, class... S>
+    struct xview_shape_type<std::array<I, L>, S...>
+    {
+        using type = std::array<I, L - integral_count<S...>()>;
+    };
+
     /************************
      * xview implementation *
      ************************/
@@ -276,21 +328,15 @@ namespace xt
     template <class E, class... S>
     template <class... SL>
     inline xview<E, S...>::xview(E& e, SL&&... slices) noexcept
-        : m_e(e), m_slices(std::forward<SL>(slices)...)
+        : m_e(e), m_slices(std::forward<SL>(slices)...),
+          m_shape(make_sequence<shape_type>(m_e.dimension() - integral_count<S...>(), 0))
     {
-        auto func = [](const auto& s) { return get_size(s); };
-        m_shape.resize(dimension());
+        auto func = [](const auto& s) noexcept { return get_size(s); };
         for (size_type i = 0; i != dimension(); ++i)
         {
             size_type index = integral_skip<S...>(i);
-            if (index < sizeof...(S))
-            {
-                m_shape[i] = apply<std::size_t>(index, func, m_slices);
-            }
-            else
-            {
-                m_shape[i] = m_e.shape()[index];
-            }
+            m_shape[i] = index < sizeof...(S) ?
+                apply<std::size_t>(index, func, m_slices) : m_e.shape()[index];
         }
     }
     //@}
@@ -359,6 +405,41 @@ namespace xt
         return access_impl(std::make_index_sequence<sizeof...(Args) + integral_count<S...>()>(), args...);
     }
 
+    template <class E, class... S>
+    inline auto xview<E, S...>::operator[](const xindex& index) -> reference
+    {
+        return element(index.cbegin(), index.cend());
+    }
+
+    template <class E, class... S>
+    template <class It>
+    inline auto xview<E, S...>::element(It first, It last) -> reference
+    {
+        auto index = make_sequence<base_index_type>(m_e.dimension(), 0);
+        auto func1 = [&first](const auto& s)
+        {
+            return get_slice_value(s, first);
+        };
+        auto func2 = [](const auto& s)
+        {
+            return xt::value(s, 0);
+        };
+        for (size_type i=0; i!=m_e.dimension(); ++i)
+        {
+            if (first != last)
+            {
+                index[i] = i < sizeof...(S) ?
+                    apply<size_type>(i, func1, m_slices) : *first++;
+            }
+            else
+            {
+                index[i] = i < sizeof...(S) ?
+                    apply<size_type>(i, func2, m_slices) : 0;
+            }
+        }
+        return m_e.element(index.cbegin(), index.cend());
+    }
+
     /**
      * Returns a constant reference to the element at the specified position in the view.
      * @param args a list of indices specifying the position in the view. Indices must be
@@ -371,6 +452,41 @@ namespace xt
     {
         return access_impl(std::make_index_sequence<sizeof...(Args) + integral_count<S...>()>(), args...);
     }
+    
+    template <class E, class... S>
+    inline auto xview<E, S...>::operator[](const xindex& index) const -> const_reference
+    {
+        return element(index.cbegin(), index.cend());
+    }
+
+    template <class E, class... S>
+    template <class It>
+    inline auto xview<E, S...>::element(It first, It last) const -> const_reference
+    {
+        auto index = make_sequence<typename E::shape_type>(m_e.dimension(), 0);
+        auto func1 = [&first](const auto& s)
+        {
+            return get_slice_value(s, first);
+        };
+        auto func2 = [](const auto& s)
+        {
+            return xt::value(s, 0);
+        };
+        for (size_type i=0; i!=m_e.dimension(); ++i)
+        {
+            if (first != last)
+            {
+                index[i] = i < sizeof...(S) ?
+                    apply<size_type>(i, func1, m_slices) : *first++;
+            }
+            else
+            {
+                index[i] = i < sizeof...(S) ?
+                    apply<size_type>(i, func2, m_slices) : 0;
+            }
+        }
+        return m_e.element(index.cbegin(), index.cend());
+    }
     //@}
 
     /**
@@ -380,7 +496,7 @@ namespace xt
     /**
      * Broadcast the shape of the view to the specified parameter.
      * @param shape the result shape
-     * @return a boolean indicating whether the broadcast is trivial
+     * @return a boolean indicating whether the broadcasting is trivial
      */
     template <class E, class... S>
     template <class ST>
@@ -390,13 +506,13 @@ namespace xt
     }
 
     /**
-     * Compares the specified strides with those of the view to see wether
-     * the broadcast is trivial.
-     * @return a boolean indicating whether the broadcast is trivial
+     * Compares the specified strides with those of the view to see whether
+     * the broadcasting is trivial.
+     * @return a boolean indicating whether the broadcasting is trivial
      */
     template <class E, class... S>
     template <class ST>
-    inline bool xview<E, S...>::is_trivial_broadcast(const ST& strides) const
+    inline bool xview<E, S...>::is_trivial_broadcast(const ST& /*strides*/) const
     {
         return false;
     }
@@ -431,10 +547,17 @@ namespace xt
     }
 
     template <class E, class... S>
-    template<typename E::size_type I, class T, class... Args>
-    inline auto xview<E, S...>::sliced_access(const xslice<T>& slice, Args... args) const -> size_type
+    template<typename E::size_type I, class T>
+    inline auto xview<E, S...>::sliced_access(const xslice<T>& slice) const -> size_type
     {
-        return slice.derived_cast()(argument<I>(args...));
+        return slice.derived_cast()(0);
+    }
+
+    template <class E, class... S>
+    template<typename E::size_type I, class T, class Arg, class... Args>
+    inline auto xview<E, S...>::sliced_access(const xslice<T>& slice, Arg arg, Args... args) const -> size_type
+    {
+        return slice.derived_cast()(argument<I>(arg, args...));
     }
 
     template <class E, class... S>
@@ -447,7 +570,7 @@ namespace xt
     template <class E, class... S>
     inline void xview<E, S...>::assign_temporary_impl(temporary_type& tmp)
     {
-        std::copy(tmp.storage_begin(), tmp.storage_end(), begin());
+        std::copy(tmp.storage_cbegin(), tmp.storage_cend(), begin());
     }
 
     namespace detail
@@ -456,9 +579,9 @@ namespace xt
         inline xview<E, get_slice_type<E, S>...> make_view_impl(E& e, std::index_sequence<I...>, S&&... slices)
         {
             return xview<E, get_slice_type<E, S>...>(
-                    e,
-                    std::forward<get_slice_type<E, S>>(get_slice_implementation(e, slices, I))...
-                    );
+                e,
+                std::forward<get_slice_type<E, S>>(get_slice_implementation(e, slices, I))...
+            );
         }
     }
 
@@ -548,9 +671,10 @@ namespace xt
      * @param shape the shape used for braodcasting
      */
     template <class E, class... S>
-    inline auto xview<E, S...>::xbegin(const shape_type& shape) -> iterator
+    template <class ST>
+    inline auto xview<E, S...>::xbegin(const ST& shape) -> xiterator<stepper, ST>
     {
-        return iterator(stepper_begin(shape), shape);
+        return xiterator<stepper, ST>(stepper_begin(shape), shape);
     }
 
     /**
@@ -559,9 +683,10 @@ namespace xt
      * @param shape the shape used for broadcasting
      */
     template <class E, class... S>
-    inline auto xview<E, S...>::xend(const shape_type& shape) -> iterator
+    template <class ST>
+    inline auto xview<E, S...>::xend(const ST& shape) -> xiterator<stepper, ST>
     {
-        return iterator(stepper_end(shape), shape);
+        return xiterator<stepper, ST>(stepper_end(shape), shape);
     }
 
     /**
@@ -570,9 +695,10 @@ namespace xt
      * @param shape the shape used for braodcasting
      */
     template <class E, class... S>
-    inline auto xview<E, S...>::xbegin(const shape_type& shape) const -> const_iterator
+    template <class ST>
+    inline auto xview<E, S...>::xbegin(const ST& shape) const -> xiterator<const_stepper, ST>
     {
-        return const_iterator(stepper_begin(shape), shape);
+        return xiterator<const_stepper, ST>(stepper_begin(shape), shape);
     }
 
     /**
@@ -581,9 +707,10 @@ namespace xt
      * @param shape the shape used for broadcasting
      */
     template <class E, class... S>
-    inline auto xview<E, S...>::xend(const shape_type& shape) const -> const_iterator
+    template <class ST>
+    inline auto xview<E, S...>::xend(const ST& shape) const -> xiterator<const_stepper, ST>
     {
-        return const_iterator(stepper_end(shape), shape);
+        return xiterator<const_stepper, ST>(stepper_end(shape), shape);
     }
 
     /**
@@ -592,7 +719,8 @@ namespace xt
      * @param shape the shape used for braodcasting
      */
     template <class E, class... S>
-    inline auto xview<E, S...>::cxbegin(const shape_type& shape) const -> const_iterator
+    template <class ST>
+    inline auto xview<E, S...>::cxbegin(const ST& shape) const -> xiterator<const_stepper, ST>
     {
         return xbegin(shape);
     }
@@ -603,7 +731,8 @@ namespace xt
      * @param shape the shape used for broadcasting
      */
     template <class E, class... S>
-    inline auto xview<E, S...>::cxend(const shape_type& shape) const -> const_iterator
+    template <class ST>
+    inline auto xview<E, S...>::cxend(const ST& shape) const -> xiterator<const_stepper, ST>
     {
         return xend(shape);
     }
@@ -682,7 +811,7 @@ namespace xt
     template <class E, class... S>
     inline auto xview<E, S...>::storage_begin() const -> const_storage_iterator
     {
-        return begin();
+        return cbegin();
     }
 
     /**
@@ -692,7 +821,28 @@ namespace xt
     template <class E, class... S>
     inline auto xview<E, S...>::storage_end() const -> const_storage_iterator
     {
-        return end();
+        return cend();
+    }
+    //@}
+
+    /**
+     * Returns a constant iterator to the first element of the buffer
+     * containing the elements of the view.
+     */
+    template <class E, class... S>
+    inline auto xview<E, S...>::storage_cbegin() const -> const_storage_iterator
+    {
+        return cbegin();
+    }
+
+    /**
+     * Returns a constant iterator to the element following the last
+     * element of the buffer containing the elements of the view.
+     */
+    template <class E, class... S>
+    inline auto xview<E, S...>::storage_cend() const -> const_storage_iterator
+    {
+        return cend();
     }
     //@}
 
@@ -700,14 +850,14 @@ namespace xt
      * xview_stepper implementation *
      ********************************/
 
-    template <class E, class... S>
-    inline xview_stepper<E, S...>::xview_stepper(view_type* view, substepper_type it,
-                                                 size_type offset, bool end)
+    template <bool is_const, class E, class... S>
+    inline xview_stepper<is_const, E, S...>::xview_stepper(view_type* view, substepper_type it,
+                                                           size_type offset, bool end)
         : p_view(view), m_it(it), m_offset(offset)
     {
         if(!end)
         {
-            auto func = [](const auto& s) { return xt::first_value(s); };
+            auto func = [](const auto& s) { return xt::value(s, 0); };
             for(size_type i = 0; i < sizeof...(S); ++i)
             {
                 size_type s = apply<size_type>(i, func, p_view->slices());
@@ -716,18 +866,18 @@ namespace xt
         }
     }
 
-    template <class E, class... S>
-    inline auto xview_stepper<E, S...>::operator*() const -> reference
+    template <bool is_const, class E, class... S>
+    inline auto xview_stepper<is_const, E, S...>::operator*() const -> reference
     {
         return *m_it;
     }
 
-    template <class E, class... S>
-    inline void xview_stepper<E, S...>::step(size_type dim, size_type n)
+    template <bool is_const, class E, class... S>
+    inline void xview_stepper<is_const, E, S...>::step(size_type dim, size_type n)
     {
         if(dim >= m_offset)
         {
-            auto func = [](const auto& s) { return step_size(s); };
+            auto func = [](const auto& s) noexcept { return step_size(s); };
             size_type index = integral_skip<S...>(dim);
             size_type step_size = index < sizeof...(S) ?
                 apply<size_type>(index, func, p_view->slices()) : 1;
@@ -735,12 +885,12 @@ namespace xt
         }
     }
 
-    template <class E, class... S>
-    inline void xview_stepper<E, S...>::step_back(size_type dim, size_type n)
+    template <bool is_const, class E, class... S>
+    inline void xview_stepper<is_const, E, S...>::step_back(size_type dim, size_type n)
     {
         if(dim >= m_offset)
         {
-            auto func = [](const auto& s) { return step_size(s); };
+            auto func = [](const auto& s) noexcept { return step_size(s); };
             size_type index = integral_skip<S...>(dim);
             size_type step_size = index < sizeof...(S) ?
                 apply<size_type>(index, func, p_view->slices()) : 1;
@@ -748,13 +898,13 @@ namespace xt
         }
     }
 
-    template <class E, class... S>
-    inline void xview_stepper<E, S...>::reset(size_type dim)
+    template <bool is_const, class E, class... S>
+    inline void xview_stepper<is_const, E, S...>::reset(size_type dim)
     {
         if(dim >= m_offset)
         {
-            auto size_func = [](const auto& s) { return get_size(s); };
-            auto step_func = [](const auto& s) { return step_size(s); };
+            auto size_func = [](const auto& s) noexcept { return get_size(s); };
+            auto step_func = [](const auto& s) noexcept { return step_size(s); };
             size_type index = integral_skip<S...>(dim);
             size_type size = index < sizeof...(S) ?
                 apply<size_type>(index, size_func, p_view->slices()) : p_view->shape()[dim];
@@ -765,28 +915,28 @@ namespace xt
         }
     }
 
-    template <class E, class... S>
-    inline void xview_stepper<E, S...>::to_end()
+    template <bool is_const, class E, class... S>
+    inline void xview_stepper<is_const, E, S...>::to_end()
     {
         m_it.to_end();
     }
 
-    template <class E, class... S>
-    inline bool xview_stepper<E, S...>::equal(const xview_stepper& rhs) const
+    template <bool is_const, class E, class... S>
+    inline bool xview_stepper<is_const, E, S...>::equal(const xview_stepper& rhs) const
     {
         return p_view == rhs.p_view && m_it == rhs.m_it && m_offset == rhs.m_offset;
     }
 
-    template <class E, class... S>
-    inline bool operator==(const xview_stepper<E, S...>& lhs,
-                           const xview_stepper<E, S...>& rhs)
+    template <bool is_const, class E, class... S>
+    inline bool operator==(const xview_stepper<is_const, E, S...>& lhs,
+                           const xview_stepper<is_const, E, S...>& rhs)
     {
         return lhs.equal(rhs);
     }
 
-    template <class E, class... S>
-    inline bool operator!=(const xview_stepper<E, S...>& lhs,
-                           const xview_stepper<E, S...>& rhs)
+    template <bool is_const, class E, class... S>
+    inline bool operator!=(const xview_stepper<is_const, E, S...>& lhs,
+                           const xview_stepper<is_const, E, S...>& rhs)
     {
         return !(lhs.equal(rhs));
     }
